# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the “License”).
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import subprocess as sp
import argparse as ap
import os
import time

import boto3
import typing

NOTIFICATION_SUFFIX = "push-endpoint"


SERVICE_ROLE_ASSUME_POLICY_TEMPLATE = """{{
    "Version": "2012-10-17",
    "Statement": [
        {{
            "Effect": "Allow",
            "Principal": {{
                "Federated": "arn:aws:iam::{0}:oidc-provider/{1}"
            }},
            "Action": "sts:AssumeRoleWithWebIdentity",
            "Condition": {{
                "StringEquals": {{
                    "{1}:sub": "system:serviceaccount:{2}:{3}"
                }}
            }}
        }}
    ]
}}"""


SERVICE_ROLE_PERMISSIONS_POLICY_TEMPLATE = """{{
    "Version": "2012-10-17",
    "Statement": [
        {{
            "Effect": "Allow",
            "Action": [
                "sqs:SendMessage"
            ],
            "Resource": ["{0}"]
        }}
    ]
}}"""


QUEUE_URL_TEMPLATE = """
            - name: QUEUE_NUM_{0}
              value: {1}"""

ALLOWED_ORIGINS_TEMPLATE = """
      - prefix: {0}"""

SERVICE_ACCOUNT_NAME = "push-endpoint-sa"
PUSH_ENDPOINT_POD_NAME = "push-endpoint"
PUSH_ENDPOINT_PATH = "/api/push-endpoint"
PUSH_ENDPOINT_K8S_TEMPLATE = """
apiVersion: v1
kind: Namespace
metadata:
  labels:
    istio-injection: enabled
  name: {push_namespace}
spec:
---
apiVersion: v1
kind: ServiceAccount
metadata:
  annotations:
    eks.amazonaws.com/role-arn: {role_arn}
  name: {name}
  namespace: {push_namespace}
---
apiVersion: v1
kind: Service
metadata:
  labels:
    service: {name}
  name: {name}
  namespace: {push_namespace}
spec:
  ports:
  - name: http
    port: 8080
    protocol: TCP
    targetPort: 8080
  selector:
    app.kubernetes.io/instance: {name}
    app.kubernetes.io/name: {name}
  type: ClusterIP
---
apiVersion: apps/v1
kind: Deployment
metadata:
  labels:
    app.kubernetes.io/instance: {name}
    app.kubernetes.io/name: {name}
  name: {name}
  namespace: {push_namespace}
spec:
  selector:
    matchLabels:
      app.kubernetes.io/instance: {name}
      app.kubernetes.io/name: {name}
  template:
    metadata:
      annotations: null
      labels:
        app.kubernetes.io/instance: {name}
        app.kubernetes.io/name: {name}
    spec:
      automountServiceAccountToken: true
      containers:
        - env:
            - name: HMAC_SECRET
              value: "{hmac_secret}"
            - name: AWS_REGION
              value: "{aws_region}"
            - name: MAX_QUEUE_NUM
              value: "{num_queues}"
{queue_urls}
          image: "{image}"
          imagePullPolicy: Always
          name: {name}
          ports:
            - containerPort: 8080
              name: http
              protocol: TCP
          resources:
            limits:
              memory: 900M
            requests:
              cpu: 1000m
              memory: 900M
          securityContext:
            allowPrivilegeEscalation: false
            capabilities:
              drop:
                - ALL
            readOnlyRootFilesystem: true
            runAsNonRoot: true
            runAsUser: 10001
          volumeMounts:
            - mountPath: /tmp
              name: tmp-volume
      securityContext:
        fsGroup: 1337
        seccompProfile:
          type: RuntimeDefault
      serviceAccountName: {name}
      volumes:
        - emptyDir:
            sizeLimit: 2Gi
          name: tmp-volume
---
apiVersion: security.istio.io/v1beta1
kind: AuthorizationPolicy
metadata:
  name: {name}-allow-services
  namespace: {push_namespace}
spec:
  action: ALLOW
  rules:
  - from:
    - source:
        principals:
        - cluster.local/ns/{ingress_namespace}/sa/{ingress_gateway_name}
        - cluster.local/ns/{core_namespace}/sa/os-notification
        - cluster.local/ns/{core_namespace}/sa/os-register
  selector:
    matchLabels:
      app.kubernetes.io/name: {name}
---
apiVersion: networking.istio.io/v1alpha3
kind: DestinationRule
metadata:
  name: {name}
  namespace: {push_namespace}
spec:
  host: {name}
  trafficPolicy:
    portLevelSettings:
    - connectionPool:
        http:
          http1MaxPendingRequests: 10000
          http2MaxRequests: 100
        tcp:
          maxConnections: 0
      port:
        number: 8080
      tls:
        mode: ISTIO_MUTUAL
---
apiVersion: networking.istio.io/v1alpha3
kind: VirtualService
metadata:
  name: {name}
  namespace: {push_namespace}
spec:
  gateways:
  - {ingress_namespace}/{osdu_gateway_name}
  hosts:
  - '*'
  http:
  - corsPolicy:
      allowCredentials: true
      allowHeaders:
      - Authorization
      - Data-Partition-Id
      - Correlation-Id
      - Content-Type
      allowMethods:
      - POST
      - GET
      - PATCH
      - PUT
      - DELETE
      allowOrigins:
{allowed_origins}
      maxAge: 60m
    headers:
      response:
        remove:
        - access-control-allow-origin
        - access-control-allow-credentials
        set:
          access-control-allow-headers: Authorization, Data-Partition-Id, Correlation-Id,
            Content-Type
          access-control-allow-methods: POST, GET, PATCH, PUT, DELETE
    match:
    - uri:
        prefix: {api_path}
    route:
    - destination:
        host: {name}
        port:
          number: 8080
"""

class CommonParameters:
    def __init__(self, osdu_instance: str, eks_cluster: str, region: str,
        num_queues: int, env_file: str, base_url: str, domain_name: str,
        endpoint_suffix: str, kube_file: typing.Optional[str], api_path: str):
        self.account_id = boto3.client('sts').get_caller_identity().get('Account')
        self.osdu_instance = osdu_instance
        self.eks_cluster = eks_cluster
        self.region = region
        self.domain_name = domain_name
        self.endpoint_suffix = endpoint_suffix
        self.kube_file = kube_file
        self.namespace_prefix: str = ""
        self.ingress_namespace: str = ""
        self.ingress_gateway: str = ""
        self.osdu_gateway: str = ""
        self.amp_workspace: str = ""
        self.num_queues = num_queues
        self.env_file = env_file
        self.base_url = base_url
        self.sa_role_policy_name = "int-test-notification-policy"
        self._set_parameters()
        self.api_path = api_path
        self.wildcard_path = f"{self.api_path}/*"
        self.sqs_client = boto3.client('sqs', region_name=region)
        self.iam_client = boto3.client('iam', region_name=region)

        if kube_file is None:
            self.kube_file = f"{self.push_namespace}.yaml"

    def _set_parameters(self):
        sns_client = boto3.client('ssm', region_name=self.region)
        parameters = {
            f"/osdu/eks/{self.eks_cluster}/amp/workspace/id": "amp_workspace",
            f"/osdu/eks/{self.eks_cluster}/instances/{self.osdu_instance}/ingress/osdu-gateway/name": "osdu_gateway",
            f"/osdu/eks/{self.eks_cluster}/instances/{self.osdu_instance}/ingress/ingress-gateway/name": "ingress_gateway",
            f"/osdu/eks/{self.eks_cluster}/instances/{self.osdu_instance}/ingress/namespace": "ingress_namespace",
            f"/osdu/instances/{self.osdu_instance}/config/k8s/namespace-prefix": "namespace_prefix",
        }
        params = sns_client.get_parameters(Names=[key for key in parameters])
        for parameter in params["Parameters"]:
            parameter_path = parameter["Name"]
            short_name = parameters[parameter_path]
            setattr(self, short_name, parameter["Value"])

        self.core_namespace = self.get_namespace("core")
        self.push_namespace = self.get_namespace(self.endpoint_suffix)
        print("Push Endpoint namespace:", self.push_namespace)

    def get_namespace(self, suffix: str) -> str:
        return f"{self.namespace_prefix}-{suffix}"

    def get_service_role(self) -> str:
        return f"{self.region}-{self.push_namespace}"

    def get_sqs_queue_name(self, queue_num: int) -> str:
        return f"{self.push_namespace}-{queue_num}"

    def get_oidc_id(self) -> str:
        eks_client = boto3.client('eks', region_name=self.region)
        cluster_info = eks_client.describe_cluster(name=self.eks_cluster)
        return cluster_info["cluster"]["identity"]["oidc"]["issuer"] \
            [len("https://"):]

    def create_policy_document(self, queue_arns: typing.List[str]) -> str:
        service_role_name = self.get_service_role()
        oidc_id = self.get_oidc_id()
        assume_account_policy = SERVICE_ROLE_ASSUME_POLICY_TEMPLATE \
            .format(self.account_id, oidc_id, self.push_namespace,
                    PUSH_ENDPOINT_POD_NAME)
        permissions_policy = SERVICE_ROLE_PERMISSIONS_POLICY_TEMPLATE \
            .format('", "'.join(queue_arns))
        role_arn: str
        try:
            role = self.iam_client.create_role(RoleName=service_role_name,
                                               AssumeRolePolicyDocument=assume_account_policy)
            self.iam_client.put_role_policy(RoleName=service_role_name,
                                            PolicyName=self.sa_role_policy_name,
                                            PolicyDocument=permissions_policy)
            role_arn = role["Role"]["Arn"]
            print(f"Created role {role_arn}")
        except self.iam_client.exceptions.EntityAlreadyExistsException:
            role = self.iam_client.get_role(RoleName=service_role_name)
            role_arn = role["Role"]["Arn"]
            print(f"Using pre-existing role {role_arn}")
        return role_arn

    def create_k8s_file(self, queue_urls: typing.List[str],
        env_file: typing.TextIO, role_arn: str):
        queue_url_templates = []
        for i, url in enumerate(queue_urls):
            queue_url_templates.append(QUEUE_URL_TEMPLATE.format(i, url))
            env_file.write(f"export PUSH_ENDPOINT_QUEUE_{i}={url}\n")
            env_file.write(f"export PUSH_ENDPOINT_URL_{i}={self.base_url}{self.api_path}/{i}\n")
        env_file.write(f"export PUSH_ENDPOINT_NUM_QUEUES={self.num_queues}\n")
        queue_url_template = "".join(queue_url_templates)[1:]

        hmac_secret = sp.check_output(["openssl", "rand", "-hex", "15"]).decode("utf-8").strip()

        notification_command = f'kubectl get deployment -n {self.core_namespace} os-notification --output json | jq -r ".spec.template.spec.containers[0].image"'
        notification_result = sp.run(notification_command, shell=True, capture_output=True, encoding="utf-8")
        if notification_result.stderr:
            print("Notification command had standard error:", notification_result.stderr)

        notification_image = notification_result.stdout.strip()
        image_parts = notification_image.split(":")
        print(image_parts, notification_image)
        if len(image_parts) != 2:
            raise RuntimeError("Unexpected result from Kubectl command.")

        # Depending on build env, images names could have two formats:
        # - `build.dev.1101.COMMIT_HASH` on CodePipeline
        # - `COMMIT_HASH` on Gitlab
        # Using -1 as an index here after the split avoid an IndexError with the second case
        endpoint_image_tag = f"push_endpoint.{image_parts[1].split('.', 1)[-1]}"
        endpoint_image = f"{image_parts[0]}:{endpoint_image_tag[:120]}"
        env_file.write(f"export HMAC_SECRET={hmac_secret}\n")

        allowed_origins = [
            ALLOWED_ORIGINS_TEMPLATE.format("http://localhost"),
            ALLOWED_ORIGINS_TEMPLATE.format(f"https://{self.domain_name}")
        ]
        print(allowed_origins, endpoint_image)

        values_contents = PUSH_ENDPOINT_K8S_TEMPLATE \
            .format(core_namespace=self.core_namespace,
                    ingress_namespace=self.ingress_namespace,
                    push_namespace=self.push_namespace,
                    ingress_gateway_name=self.ingress_gateway,
                    osdu_gateway_name=self.osdu_gateway,
                    name=PUSH_ENDPOINT_POD_NAME,
                    hmac_secret=hmac_secret,
                    aws_region=self.region,
                    num_queues=self.num_queues,
                    image=endpoint_image,
                    queue_urls=queue_url_template,
                    api_path=self.api_path,
                    allowed_origins="".join(allowed_origins)[1:],
                    role_arn=role_arn)

        with open(self.kube_file, "w") as file:
            file.write(values_contents)

    async def ensure_queue_created(self, queue_num: int,
        received_queues: asyncio.Queue):
        queue_name = self.get_sqs_queue_name(queue_num)
        try:
            queue = self.sqs_client.create_queue(QueueName=queue_name)
        except self.sqs_client.exceptions.QueueDeletedRecently:
            time.sleep(62)
            await self.ensure_queue_created(queue_num, received_queues)
            return
        except self.sqs_client.exceptions.QueueNameExists:
            queue = self.sqs_client.get_queue_url(QueueName=queue_name)
        await received_queues.put(queue["QueueUrl"])

    def get_sqs_queue_arn(self, queue_url: str):
        QUEUE_ARN_ATTR = "QueueArn"
        response = self.sqs_client.get_queue_attributes(
            QueueUrl=queue_url,
            AttributeNames=[QUEUE_ARN_ATTR]
        )
        return response["Attributes"][QUEUE_ARN_ATTR]

    def disable_jwt_auth_on_path(self):
        output = sp.check_output(["kubectl", "get", "authorizationpolicy", "-n",
                                  self.ingress_namespace, "ingress-jwt-required",
                                  "-o", "json"])
        current_auth_policy = json.loads(output)
        for rule in current_auth_policy["spec"]["rules"]:
            for outer_op in rule["to"]:
                not_paths = outer_op["operation"]["notPaths"]
                if self.wildcard_path not in not_paths:
                    not_paths.append(self.wildcard_path)
        new_auth_policy = json.dumps(current_auth_policy)
        sp.check_output(["kubectl", "apply", "-f", "-"], input=new_auth_policy,
                        encoding='utf-8')

    async def create_resources(self):
        received_queues = asyncio.Queue()
        tasks = []
        for i in range(self.num_queues):
            task = asyncio.create_task(
                self.ensure_queue_created(i, received_queues)
            )
            tasks.append(task)

        await asyncio.gather(*tasks)
        print("Should have created the queues.")

        queue_arns = []
        queue_urls = []
        while not received_queues.empty():
            queue_url = await received_queues.get()
            queue_arns.append(self.get_sqs_queue_arn(queue_url))
            queue_urls.append(queue_url)

        role_arn = self.create_policy_document(queue_arns)
        print("Should have created the IAM Role and policy. ARN:", role_arn)

        with open(self.env_file, "w") as file:
            self.create_k8s_file(queue_urls, file, role_arn)

        print("Now creating the Kubernetes resources")
        sp.check_output(["kubectl", "apply", "-f", self.kube_file])

        print("Now changing the JWT Authorization Policy")
        self.disable_jwt_auth_on_path()

    async def ensure_queue_deleted(self, queue_num: int):
        queue_name = self.get_sqs_queue_name(queue_num)
        try:
            queue_url = self.sqs_client.get_queue_url(QueueName=queue_name)
            self.sqs_client.delete_queue(QueueUrl=queue_url["QueueUrl"])
        except self.sqs_client.exceptions.QueueDoesNotExist:
            print("Queue", queue_name, "not found, so could not delete!")

    def re_enable_jwt_auth_on_path(self):
        output = sp.check_output(["kubectl", "get", "authorizationpolicy", "-n",
                                  self.ingress_namespace, "ingress-jwt-required",
                                  "-o", "json"])
        current_auth_policy = json.loads(output)
        for rule in current_auth_policy["spec"]["rules"]:
            for outer_op in rule["to"]:
                not_paths = outer_op["operation"]["notPaths"]
                try:
                    not_paths.remove(self.wildcard_path)
                except ValueError:
                    print(f"Path {self.wildcard_path} was already not exempted from the disable JWT authentication.")

        new_auth_policy = json.dumps(current_auth_policy)
        sp.check_output(["kubectl", "apply", "-f", "-"], input=new_auth_policy,
                        encoding='utf-8')

    async def delete_resources(self):
        print("Now changing the JWT Authorization Policy")
        self.re_enable_jwt_auth_on_path()

        print("Deleting the kubernetes resources.")
        try:
            sp.check_output(["kubectl", "delete", "-f", self.kube_file])
        except Exception as e:
            print("Failed to delete the kubernetes resources:", e)

        print("Deleting the role policy.")
        try:
            self.iam_client.delete_role_policy(RoleName=self.get_service_role(),
                                               PolicyName=self.sa_role_policy_name)
            self.iam_client.delete_role(RoleName=self.get_service_role())
        except self.iam_client.exceptions.NoSuchEntityException:
            print("Could not delete role", self.get_service_role(),
                  "because it does not exist!")

        print("Deleting the SQS queues.")
        tasks = []
        for i in range(self.num_queues):
            task = asyncio.create_task(self.ensure_queue_deleted(i))
            tasks.append(task)

        await asyncio.gather(*tasks)


async def main():
    argparser = ap.ArgumentParser(description="Handles creation/deletion of the"
                                              " AWS resources required for the integration te"
                                              "sts and creates an environment variable file f"
                                              "or the integration tests to use.")
    argparser.add_argument('--region', type=str,
                           default=os.getenv('AWS_REGION'),
                           help="The AWS region to deploy the cloud resources to.")
    argparser.add_argument('--num_queues', type=int, default=2,
                           help="The number of queues to create.")
    argparser.add_argument('--osdu_instance_name', type=str,
                           default=os.getenv("OSDU_INSTANCE_NAME"),
                           help="The OSDU instance name.")
    argparser.add_argument('--eks_cluster_name', type=str,
                           default=os.getenv("EKS_CLUSTER_NAME"))
    argparser.add_argument('--base_url', type=str,
                           default=os.getenv("AWS_BASE_URL"))
    argparser.add_argument('--domain_name', type=str,
                           default=os.getenv("DOMAIN"))
    argparser.add_argument('--endpoint_suffix', type=str,
                           default=NOTIFICATION_SUFFIX)
    argparser.add_argument('--kube_file', type=str, default=None)
    argparser.add_argument('--api_path', type=str,
                           default=PUSH_ENDPOINT_PATH,
                           help="The push endpoint path relative from the ingress gateway.")
    argparser.add_argument('--environment_file', type=str,
                           default="./environment_file.sh",
                           help="The file to create the environment variables in.")
    argparser.add_argument('--create', action='store_true', help="Creates the "
                                                                 "resources required for the integration tests.")
    argparser.add_argument('--delete', action='store_true', help="Deletes the "
                                                                 "resources created for the integration tests.")

    args = argparser.parse_args()
    if args.create and args.delete:
        raise ValueError("Cannot create and delete at the same time.")

    if not args.create and not args.delete:
        raise ValueError("Must specify either create or delete.")

    osdu_instance = args.osdu_instance_name
    eks_cluster = args.eks_cluster_name
    num_queues = args.num_queues
    region = args.region
    env_file = args.environment_file
    base_url = args.base_url
    domain_name = args.domain_name
    endpoint_suffix = args.endpoint_suffix
    kube_file = args.kube_file
    api_path = args.api_path
    common_params = CommonParameters(osdu_instance, eks_cluster, region,
                                     num_queues, env_file, base_url, domain_name,
                                     endpoint_suffix, kube_file, api_path)

    if args.create:
        await common_params.create_resources()

    if args.delete:
        await common_params.delete_resources()


if __name__ == '__main__':
    asyncio.run(main())

